"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var request = require("request");
var assert = require("assert-plus");
var is_stream_1 = require("is-stream");
var stream_1 = require("stream");
var utils_1 = require("./utils");
var chunkedUploadMinSize = 150e+6;
var chunkedUploadChunkSize = 5e+6;
var dropboxApiBasePath = 'https://content.dropboxapi.com/2/files';
function upload(options) {
    return __awaiter(this, void 0, void 0, function () {
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    // verfiy options
                    assert.object(options, 'options');
                    assert.string(options.access_token, 'options.access_token');
                    assert.ok(is_stream_1.readable(options.readable_stream), 'options.readable_stream must be a readable stream');
                    assert.number(options.file_size, 'options.file_size');
                    assert.string(options.destination, 'options.destination');
                    assert.optionalBool(options.forced_chunked_upload, 'options.forced_chunked_upload');
                    assert.optionalBool(options.mute, 'options.mute');
                    assert.optionalBool(options.autorename, 'options.autorename');
                    assert.optionalString(options.mode, 'options.mode');
                    assert.ok(options.cancel === undefined, 'options.cancel must be undefined. its set by this library');
                    options.autorename = options.autorename === undefined ? true : options.autorename;
                    options.mute = options.mute === undefined ? false : options.mute;
                    options.mode = options.mode || 'add';
                    options.chunk_size = options.chunk_size || chunkedUploadChunkSize;
                    if (!(options.forced_chunked_upload || options.file_size > chunkedUploadMinSize)) return [3 /*break*/, 2];
                    return [4 /*yield*/, handleChunkedUpload(options)];
                case 1: return [2 /*return*/, _a.sent()];
                case 2: return [4 /*yield*/, handleSimpleUpload(options)];
                case 3: return [2 /*return*/, _a.sent()];
            }
        });
    });
}
exports.default = upload;
function handleSimpleUpload(options) {
    return __awaiter(this, void 0, void 0, function () {
        return __generator(this, function (_a) {
            return [2 /*return*/, utils_1.promiseFromCallback(function (next) {
                    var args = {
                        path: options.destination,
                        mode: options.mode,
                        autorename: options.autorename,
                        mute: options.mute
                    };
                    if (options.client_modified) {
                        args.client_modified = options.client_modified;
                    }
                    var writeStream = request.post({
                        uri: dropboxApiBasePath + "/upload",
                        method: 'POST',
                        headers: {
                            'Authorization': "Bearer " + options.access_token,
                            'Content-Type': 'application/octet-stream',
                            'Dropbox-API-Arg': JSON.stringify(args),
                        }
                    }, function (err, resp, body) {
                        if (err) {
                            return next(err);
                        }
                        if (resp.statusCode !== 200) {
                            return next(utils_1.getHttpError(resp.statusCode, body));
                        }
                        return next(null, utils_1.safeParseJson(body));
                    });
                    options.cancel = function () {
                        writeStream.abort();
                        next(new Error("user_aborted"));
                    };
                    options.readable_stream.pipe(writeStream);
                })];
        });
    });
}
function handleChunkedUpload(options) {
    return __awaiter(this, void 0, void 0, function () {
        var hasUserAborted, onCancel, sessionId, doneSize, pendingData, _loop_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    hasUserAborted = false;
                    options.cancel = function () {
                        hasUserAborted = true;
                        if (onCancel) {
                            onCancel();
                        }
                    };
                    return [4 /*yield*/, uploadSessionStart(options)
                        // console.log('uploadSessionStarted', sessionId)
                    ];
                case 1:
                    sessionId = _a.sent();
                    doneSize = 0;
                    options.readable_stream.pause();
                    _loop_1 = function () {
                        var doneChunkSize, stream;
                        return __generator(this, function (_a) {
                            switch (_a.label) {
                                case 0:
                                    doneChunkSize = 0;
                                    stream = new stream_1.Readable({
                                        read: function (size) {
                                            function step(rs) {
                                                return __awaiter(this, void 0, void 0, function () {
                                                    var data, dataSmallerThanChunkSize, dataFitsInChunkUnit, dataWontCompleteFile, existingDataSmallerThanChunkSize;
                                                    return __generator(this, function (_a) {
                                                        switch (_a.label) {
                                                            case 0:
                                                                if (hasUserAborted) {
                                                                    rs.push(null);
                                                                    return [2 /*return*/];
                                                                }
                                                                if (!pendingData) return [3 /*break*/, 3];
                                                                doneChunkSize += pendingData.length;
                                                                doneSize += pendingData.length;
                                                                if (!rs.push(pendingData)) return [3 /*break*/, 3];
                                                                pendingData = false;
                                                                // can read more
                                                                return [4 /*yield*/, utils_1.nextTick()];
                                                            case 1:
                                                                // can read more
                                                                _a.sent();
                                                                return [4 /*yield*/, step(rs)];
                                                            case 2:
                                                                _a.sent();
                                                                return [2 /*return*/];
                                                            case 3:
                                                                // if file is done end the stream
                                                                if (doneSize >= options.file_size) {
                                                                    rs.push(null);
                                                                    return [2 /*return*/];
                                                                }
                                                                data = options.readable_stream.read(size);
                                                                if (!!data) return [3 /*break*/, 6];
                                                                return [4 /*yield*/, utils_1.nextTick()];
                                                            case 4:
                                                                _a.sent();
                                                                return [4 /*yield*/, step(rs)];
                                                            case 5:
                                                                _a.sent();
                                                                return [2 /*return*/];
                                                            case 6:
                                                                dataSmallerThanChunkSize = data.length < options.chunk_size;
                                                                dataFitsInChunkUnit = data.length + doneChunkSize <= options.chunk_size;
                                                                dataWontCompleteFile = data.length + doneChunkSize < options.file_size;
                                                                existingDataSmallerThanChunkSize = doneChunkSize < options.chunk_size;
                                                                if (!((!existingDataSmallerThanChunkSize) || (dataSmallerThanChunkSize && !dataFitsInChunkUnit && dataWontCompleteFile))) return [3 /*break*/, 7];
                                                                // need new stream
                                                                // - if existing data has already filled the chunk unit (!existingDataSmallerThanChunkSize)
                                                                // - or if dataSmallerThanChunkSize && !dataFitsInChunkUnit && dataWontCompleteFile
                                                                pendingData = data;
                                                                rs.push(null);
                                                                return [3 /*break*/, 10];
                                                            case 7:
                                                                // carry on with current stream
                                                                doneChunkSize += data.length;
                                                                doneSize += data.length;
                                                                if (!rs.push(data)) return [3 /*break*/, 10];
                                                                // can read more
                                                                return [4 /*yield*/, utils_1.nextTick()];
                                                            case 8:
                                                                // can read more
                                                                _a.sent();
                                                                return [4 /*yield*/, step(rs)];
                                                            case 9:
                                                                _a.sent();
                                                                return [2 /*return*/];
                                                            case 10: return [2 /*return*/];
                                                        }
                                                    });
                                                });
                                            }
                                            step(this);
                                        }
                                    });
                                    return [4 /*yield*/, uploadSessionAppend(options, sessionId, doneSize, stream, function (onCancelFn) {
                                            onCancel = onCancelFn;
                                        })];
                                case 1:
                                    _a.sent();
                                    return [2 /*return*/];
                            }
                        });
                    };
                    _a.label = 2;
                case 2:
                    if (!(!hasUserAborted && doneSize < options.file_size)) return [3 /*break*/, 4];
                    return [5 /*yield**/, _loop_1()];
                case 3:
                    _a.sent();
                    return [3 /*break*/, 2];
                case 4:
                    if (hasUserAborted) {
                        throw new Error("user_aborted");
                    }
                    return [4 /*yield*/, uploadSessionFinish(options, sessionId, doneSize)];
                case 5: return [2 /*return*/, _a.sent()];
            }
        });
    });
}
function uploadSessionStart(options) {
    return __awaiter(this, void 0, void 0, function () {
        var sessionRes;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, utils_1.makeRequest({
                        url: dropboxApiBasePath + "/upload_session/start",
                        method: 'POST',
                        headers: {
                            'Authorization': "Bearer " + options.access_token,
                            'Content-Type': 'application/octet-stream',
                            'Dropbox-API-Arg': JSON.stringify({ close: false })
                        }
                    })];
                case 1:
                    sessionRes = _a.sent();
                    assert.object(sessionRes.body, 'cannot start upload session ' + JSON.stringify(sessionRes));
                    assert.string(sessionRes.body.session_id, 'cannot start upload session ' + JSON.stringify(sessionRes));
                    return [2 /*return*/, sessionRes.body.session_id];
            }
        });
    });
}
function uploadSessionAppend(options, sessionId, doneSize, stream, registerOnCancel) {
    return __awaiter(this, void 0, void 0, function () {
        return __generator(this, function (_a) {
            return [2 /*return*/, utils_1.promiseFromCallback(function (next) {
                    var args = {
                        cursor: {
                            session_id: sessionId,
                            offset: doneSize
                        }
                    };
                    // console.log('uploadSessionAppend', args)
                    var writeStream = request.post({
                        uri: dropboxApiBasePath + "/upload_session/append_v2",
                        method: 'POST',
                        headers: {
                            'Authorization': "Bearer " + options.access_token,
                            'Content-Type': 'application/octet-stream',
                            'Dropbox-API-Arg': JSON.stringify(args),
                        }
                    }, function (err, resp, body) {
                        // console.log('uploadSessionAppend - r', err, body)
                        if (err) {
                            return next(err);
                        }
                        if (resp.statusCode !== 200) {
                            return next(utils_1.getHttpError(resp.statusCode, ''));
                        }
                        return next(null);
                    });
                    registerOnCancel(function () {
                        writeStream.abort();
                    });
                    stream.pipe(writeStream);
                })];
        });
    });
}
function uploadSessionFinish(options, sessionId, doneSize) {
    return __awaiter(this, void 0, void 0, function () {
        var args, res;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    args = {
                        cursor: {
                            session_id: sessionId,
                            offset: doneSize
                        },
                        commit: {
                            path: options.destination,
                            mode: options.mode,
                            autorename: options.autorename,
                            mute: options.mute
                        }
                    };
                    if (options.client_modified) {
                        args.commit.client_modified = options.client_modified;
                    }
                    return [4 /*yield*/, utils_1.makeRequest({
                            url: dropboxApiBasePath + "/upload_session/finish",
                            method: 'POST',
                            headers: {
                                'Authorization': "Bearer " + options.access_token,
                                'Content-Type': 'application/octet-stream',
                                'Dropbox-API-Arg': JSON.stringify(args)
                            }
                        })
                        // console.log('uploadSessionFinish - r', res)
                    ];
                case 1:
                    res = _a.sent();
                    // console.log('uploadSessionFinish - r', res)
                    assert.object(res.body, 'cannot finish upload session ' + JSON.stringify(res));
                    return [2 /*return*/, res.body];
            }
        });
    });
}
